@@ -8,9 +8,7 @@ from functools import wraps
 import getpass
 import os
 import re
-import threading
 import time
-import select
 import socket
 import sys
 
@@ -21,28 +19,39 @@ from fabric.exceptions import NetworkError
 try:
     import warnings
     warnings.simplefilter('ignore', DeprecationWarning)
-    import ssh
+    import paramiko as ssh
 except ImportError, e:
     import traceback
     traceback.print_exc()
-    print >> sys.stderr, """
+    msg = """
 There was a problem importing our SSH library (see traceback above).
 Please make sure all dependencies are installed and importable.
 """.rstrip()
+    sys.stderr.write(msg + '\n')
     sys.exit(1)
 
 
-host_pattern = r'((?P<user>.+)@)?(?P<host>[^:]+)(:(?P<port>\d+))?'
-host_regex = re.compile(host_pattern)
+ipv6_regex = re.compile('^\[?(?P<host>[0-9A-Fa-f:]+)\]?(:(?P<port>\d+))?$')
+
+
+def direct_tcpip(client, host, port):
+    return client.get_transport().open_channel(
+        'direct-tcpip',
+        (host, int(port)),
+        ('', 0)
+    )
 
 
 class HostConnectionCache(dict):
     """
     Dict subclass allowing for caching of host connections/clients.
 
-    This subclass does not offer any extra methods, but will intelligently
-    create new client connections when keys are requested, or return previously
-    created connections instead.
+    This subclass will intelligently create new client connections when keys
+    are requested, or return previously created connections instead.
+
+    It also handles creating new socket-like objects when required to implement
+    gateway connections and `ProxyCommand`, and handing them to the inner
+    connection methods.
 
     Key values are the same as host specifiers throughout Fabric: optional
     username + ``@``, mandatory hostname, optional ``:`` + port number.
@@ -71,9 +80,23 @@ class HostConnectionCache(dict):
         """
         Force a new connection to ``key`` host string.
         """
+        from fabric.state import env
         user, host, port = normalize(key)
         key = normalize_to_string(key)
-        self[key] = connect(user, host, port)
+        sock = None
+        proxy_command = ssh_config().get('proxycommand', None)
+        if env.gateway:
+            gateway = normalize_to_string(env.gateway)
+            # Ensure initial gateway connection
+            if gateway not in self:
+                self[gateway] = connect(user, host, port)
+            # Now we should have an open gw connection and can ask it for a
+            # direct-tcpip channel to the real target. (Bypass our own
+            # __getitem__ override to avoid hilarity.)
+            sock = direct_tcpip(dict.__getitem__(self, gateway), host, port)
+        elif proxy_command:
+            sock = ssh.ProxyCommand(proxy_command)
+        self[key] = connect(user, host, port, sock)
 
     def __getitem__(self, key):
         """
@@ -120,7 +143,7 @@ def ssh_config(host_string=None):
             with open(path) as fd:
                 conf.parse(fd)
                 env._ssh_config = conf
-        except IOError, e:
+        except IOError:
             abort("Unable to load SSH config file '%s'" % path)
     host = parse_host_string(host_string or env.host_string)['host']
     return env._ssh_config.lookup(host)
@@ -137,7 +160,7 @@ def key_filenames():
     from fabric.state import env
     keys = env.key_filename
     # For ease of use, coerce stringish key filename into list
-    if not isinstance(env.key_filename, (list, tuple)):
+    if isinstance(env.key_filename, basestring) or env.key_filename is None:
         keys = [keys]
     # Strip out any empty strings (such as the default value...meh)
     keys = filter(bool, keys)
@@ -150,7 +173,25 @@ def key_filenames():
 
 
 def parse_host_string(host_string):
-    return host_regex.match(host_string).groupdict()
+    # Split host_string to user (optional) and host/port
+    user_hostport = host_string.rsplit('@', 1)
+    hostport = user_hostport.pop()
+    user = user_hostport[0] if user_hostport and user_hostport[0] else None
+
+    # Split host/port string to host and optional port
+    # For IPv6 addresses square brackets are mandatory for host/port separation
+    if hostport.count(':') > 1:
+        # Looks like IPv6 address
+        r = ipv6_regex.match(hostport).groupdict()
+        host = r['host'] or None
+        port = r['port'] or None
+    else:
+        # Hostname or IPv4 address
+        host_port = hostport.rsplit(':', 1)
+        host = host_port.pop(0) or None
+        port = host_port[0] if host_port and host_port[0] else None
+
+    return {'user': user, 'host': host, 'port': port}
 
 
 def normalize(host_string, omit_port=False):
@@ -202,6 +243,7 @@ def to_dict(host_string):
         'user': user, 'host': host, 'port': port, 'host_string': host_string
     }
 
+
 def from_dict(arg):
     return join_host_strings(arg['user'], arg['host'], arg['port'])
 
@@ -213,15 +255,18 @@ def denormalize(host_string):
     If the user part is the default user, it is removed;
     if the port is port 22, it also is removed.
     """
-    from state import env
-    r = host_regex.match(host_string).groupdict()
+    from fabric.state import env
+
+    r = parse_host_string(host_string)
     user = ''
     if r['user'] is not None and r['user'] != env.user:
         user = r['user'] + '@'
     port = ''
     if r['port'] is not None and r['port'] != '22':
         port = ':' + r['port']
-    return user + r['host'] + port
+    host = r['host']
+    host = '[%s]' % host if port and host.count(':') > 1 else host
+    return user + host + port
 
 
 def join_host_strings(user, host, port=None):
@@ -231,13 +276,17 @@ def join_host_strings(user, host, port=None):
     This function is not responsible for handling missing user/port strings;
     for that, see the ``normalize`` function.
 
+    If ``host`` looks like IPv6 address, it will be enclosed in square brackets
+
     If ``port`` is omitted, the returned string will be of the form
     ``user@host``.
     """
-    port_string = ''
     if port:
-        port_string = ":%s" % port
-    return "%s@%s%s" % (user, host, port_string)
+        # Square brackets are necessary for IPv6 host/port separation
+        template = "%s@[%s]:%s" if host.count(':') > 1 else "%s@%s:%s"
+        return template % (user, host, port)
+    else:
+        return "%s@%s" % (user, host)
 
 
 def normalize_to_string(host_string):
@@ -247,9 +296,12 @@ def normalize_to_string(host_string):
     return join_host_strings(*normalize(host_string))
 
 
-def connect(user, host, port):
+def connect(user, host, port, sock=None):
     """
     Create and return a new SSHClient instance connected to given host.
+
+    If ``sock`` is given, it's passed into ``SSHClient.connect()`` directly.
+    Used for gateway connections by e.g. ``HostConnectionCache``.
     """
     from state import env, output
 
@@ -289,7 +341,8 @@ def connect(user, host, port):
                 key_filename=key_filenames(),
                 timeout=env.timeout,
                 allow_agent=not env.no_agent,
-                look_for_keys=not env.no_keys
+                look_for_keys=not env.no_keys,
+                sock=sock
             )
             connected = True
 
@@ -302,21 +355,27 @@ def connect(user, host, port):
         # command line results in the big banner error about man-in-the-middle
         # attacks.
         except ssh.BadHostKeyException, e:
-            raise NetworkError("Host key for %s did not match pre-existing key! Server's key was changed recently, or possible man-in-the-middle attack." % env.host, e)
+            raise NetworkError("Host key for %s did not match pre-existing key! Server's key was changed recently, or possible man-in-the-middle attack." % host, e)
         # Prompt for new password to try on auth failure
         except (
             ssh.AuthenticationException,
             ssh.PasswordRequiredException,
             ssh.SSHException
         ), e:
-            # For whatever reason, empty password + no ssh key or agent results
-            # in an SSHException instead of an AuthenticationException. Since
-            # it's difficult to do otherwise, we must assume empty password +
-            # SSHException == auth exception. Conversely: if we get
-            # SSHException and there *was* a password -- it is probably
-            # something non auth related, and should be sent upwards.
-            if e.__class__ is ssh.SSHException and password:
-                raise NetworkError(str(e), e)
+            msg = str(e)
+            # For whatever reason, empty password + no ssh key or agent
+            # results in an SSHException instead of an
+            # AuthenticationException. Since it's difficult to do
+            # otherwise, we must assume empty password + SSHException ==
+            # auth exception. Conversely: if we get SSHException and there
+            # *was* a password -- it is probably something non auth
+            # related, and should be sent upwards.
+            #
+            # This also holds true for rejected/unknown host keys: we have to
+            # guess based on other heuristics.
+            if e.__class__ is ssh.SSHException \
+                and (password or msg.startswith('Unknown server')):
+                raise NetworkError(msg, e)
 
             # Otherwise, assume an auth exception, and prompt for new/better
             # password.
@@ -373,7 +432,7 @@ def connect(user, host, port):
             err += ")"
             # Debuggin'
             if output.debug:
-                print >>sys.stderr, err
+                sys.stderr.write(err + '\n')
             # Having said our piece, try again
             if not giving_up:
                 # Sleep if it wasn't a timeout, so we still get timeout-like
@@ -383,14 +442,19 @@ def connect(user, host, port):
                 continue
             # Override eror msg if we were retrying other errors
             if not_timeout:
-                msg = "Low level socket error connecting to host %s: %s" % (
-                    host, e[1]
+                msg = "Low level socket error connecting to host %s on port %s: %s" % (
+                    host, port, e[1]
                 )
             # Here, all attempts failed. Tweak error msg to show # tries.
             # TODO: find good humanization module, jeez
             s = "s" if env.connection_attempts > 1 else ""
             msg += " (tried %s time%s)" % (env.connection_attempts, s)
             raise NetworkError(msg, e)
+        # Ensure that if we terminated without connecting and we were given an
+        # explicit socket, close it out.
+        finally:
+            if not connected and sock is not None:
+                sock.close()
 
 
 def prompt_for_password(prompt=None, no_colon=False, stream=None):
@@ -414,7 +478,7 @@ def prompt_for_password(prompt=None, no_colon=False, stream=None):
     handle_prompt_abort("a connection or sudo password")
     stream = stream or sys.stderr
     # Construct prompt
-    default = "[%s] Login password" % env.host_string
+    default = "[%s] Login password for '%s'" % (env.host_string, env.user)
     password_prompt = prompt if (prompt is not None) else default
     if not no_colon:
         password_prompt += ": "
@@ -455,6 +519,7 @@ def needs_host(func):
                                     " host string for connection: ")
             env.update(to_dict(host_string))
         return func(*args, **kwargs)
+    host_prompting_wrapper.undecorated = func
     return host_prompting_wrapper
 
 
@@ -469,8 +534,10 @@ def disconnect_all():
     # Explicitly disconnect from all servers
     for key in connections.keys():
         if output.status:
-            print "Disconnecting from %s..." % denormalize(key),
+            # Here we can't use the py3k print(x, end=" ")
+            # because 2.5 backwards compatibility
+            sys.stdout.write("Disconnecting from %s... " % denormalize(key))
         connections[key].close()
         del connections[key]
         if output.status:
-            print "done."
+            sys.stdout.write("done.\n")
